import asyncio

import pytest
from yarl import URL

from aio_pika import connect


VARIANTS = (
    (dict(url="amqp://localhost/"), "amqp://localhost/"),
    (dict(url="amqp://localhost"), "amqp://localhost/"),
    (dict(url="amqp://localhost:5674"), "amqp://localhost:5674/"),
    (dict(url="amqp://localhost:5674//"), "amqp://localhost:5674//"),
    (dict(url="amqp://localhost:5674/"), "amqp://localhost:5674/"),
    (dict(host="localhost", port=8888), "amqp://guest:guest@localhost:8888//"),
    (
        dict(host="localhost", port=8888, virtualhost="foo"),
        "amqp://guest:guest@localhost:8888/foo",
    ),
    (
        dict(host="localhost", port=8888, virtualhost="/foo"),
        "amqp://guest:guest@localhost:8888//foo",
    ),
)


class FakeConnection:
    def __init__(self, url, **kwargs):
        self.url = URL(url)
        self.kwargs = kwargs

    async def connect(self, timeout=None, **kwargs):
        return


@pytest.mark.parametrize("kwargs,expected", VARIANTS)
def test_simple(kwargs, expected):
    loop = asyncio.get_event_loop()
    conn: FakeConnection = loop.run_until_complete(
        connect(connection_class=FakeConnection, **kwargs),  # type: ignore
    )

    assert conn.url == URL(expected)



from typing import Any

import shortuuid


def get_random_name(*args: Any) -> str:
    prefix = ["test"]
    for item in args:
        prefix.append(item)
    prefix.append(shortuuid.uuid())

    return ".".join(prefix)



import atexit
import asyncio
import gc
import socket
import tracemalloc
from contextlib import suppress
from functools import partial
from time import sleep
from typing import Any, Callable, Generator

import aiormq
import pamqp
import pytest
from aiomisc import awaitable
from yarl import URL

import aio_pika
from .docker_client import (
    ContainerInfo,
    DockerClient,
    DockerHostInfo,
    DockerNotAvailableError,
    check_docker_available,
)

# Cached docker host info from pytest_configure
_docker_host_info: DockerHostInfo | None = None

# Global registry for atexit cleanup
_docker_client: DockerClient | None = None
_docker_containers: set[str] = set()


def _atexit_kill_containers() -> None:
    """Kill all containers on exit (handles crashes/interrupts)."""
    if _docker_client is None:
        return
    for container_id in _docker_containers:
        with suppress(Exception):
            _docker_client.kill(container_id)
        with suppress(Exception):
            _docker_client.remove(container_id)
    _docker_containers.clear()


atexit.register(_atexit_kill_containers)


def pytest_configure(config: pytest.Config) -> None:
    """Check Docker availability before running tests."""
    global _docker_host_info
    try:
        _docker_host_info = check_docker_available()
    except DockerNotAvailableError as e:
        raise pytest.UsageError(str(e)) from e


@pytest.fixture
async def add_cleanup(event_loop):
    entities = []

    def payload(func, *args, **kwargs):
        func = partial(awaitable(func), *args, **kwargs)
        entities.append(func)

    try:
        yield payload
    finally:
        for func in entities[::-1]:
            await func()

        entities.clear()


@pytest.fixture
async def create_task(event_loop):
    tasks = []

    def payload(coroutine):
        task = event_loop.create_task(coroutine)
        tasks.append(task)
        return task

    try:
        yield payload
    finally:
        cancelled = []
        for task in tasks:
            if task.done():
                continue
            task.cancel()
            cancelled.append(task)

        results = await asyncio.gather(*cancelled, return_exceptions=True)

        for result in results:
            if not isinstance(result, asyncio.CancelledError):
                raise result


@pytest.fixture(scope="session")
def docker() -> Generator[Callable[..., ContainerInfo], Any, Any]:
    global _docker_client
    _docker_client = DockerClient(_docker_host_info)

    def docker_run(
        image: str, ports: list[str],
        environment: dict[str, str] | None = None,
    ) -> ContainerInfo:
        info = _docker_client.run(image, ports, environment=environment)
        _docker_containers.add(info.id)
        return info

    try:
        yield docker_run
    finally:
        for container_id in list(_docker_containers):
            with suppress(Exception):
                _docker_client.kill(container_id)
            with suppress(Exception):
                _docker_client.remove(container_id)
            _docker_containers.discard(container_id)


@pytest.fixture(scope="session")
def rabbitmq_container(docker) -> ContainerInfo:
    info = docker("mosquito/aiormq-rabbitmq", ["5672/tcp", "5671/tcp"])
    # Readiness probe - wait for RabbitMQ to be ready
    while True:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            try:
                sock.connect((info.host, info.ports["5672/tcp"]))
                sock.send(b"AMQP\x00\x00\x09\x01")
                data = sock.recv(4)
                if len(data) == 4:
                    return info
            except ConnectionError:
                pass
        sleep(0.3)


@pytest.fixture(scope="session")
def amqp_direct_url(rabbitmq_container: ContainerInfo) -> URL:
    return URL.build(
        scheme="amqp", user="guest", password="guest", path="//",
        host=rabbitmq_container.host,
        port=rabbitmq_container.ports["5672/tcp"],
    )


@pytest.fixture
def amqp_url(request, amqp_direct_url) -> URL:
    query = dict(amqp_direct_url.query)
    query["name"] = request.node.nodeid
    return amqp_direct_url.with_query(**query)


@pytest.fixture(
    scope="module",
    params=[aio_pika.connect, aio_pika.connect_robust],
    ids=["connect", "connect_robust"],
)
def connection_fabric(request):
    return request.param


@pytest.fixture
def create_connection(connection_fabric, event_loop, amqp_url):
    return partial(connection_fabric, amqp_url, loop=event_loop)


@pytest.fixture
def create_channel(connection: aio_pika.Connection, add_cleanup):
    conn = connection

    async def fabric(cleanup=True, connection=None, *args, **kwargs):
        if connection is None:
            connection = conn

        channel = await connection.channel(*args, **kwargs)
        if cleanup:
            add_cleanup(channel.close)

        return channel

    return fabric


# noinspection PyTypeChecker
@pytest.fixture
async def connection(create_connection) -> aio_pika.Connection:  # type: ignore
    async with await create_connection() as conn:
        yield conn


# noinspection PyTypeChecker
@pytest.fixture
async def channel(      # type: ignore
    connection: aio_pika.Connection,
) -> aio_pika.Channel:
    async with connection.channel() as ch:
        yield ch


@pytest.fixture
def declare_queue(connection, channel, add_cleanup):
    ch = channel

    async def fabric(
        *args, cleanup=True, channel=None, **kwargs,
    ) -> aio_pika.Queue:
        if channel is None:
            channel = ch

        queue = await channel.declare_queue(*args, **kwargs)

        if cleanup and not kwargs.get("auto_delete"):
            add_cleanup(queue.delete)

        return queue

    return fabric


@pytest.fixture
def declare_exchange(connection, channel, add_cleanup):
    ch = channel

    async def fabric(
        *args, channel=None, cleanup=True, **kwargs,
    ) -> aio_pika.Exchange:
        if channel is None:
            channel = ch

        exchange = await channel.declare_exchange(*args, **kwargs)

        if cleanup and not kwargs.get("auto_delete"):
            add_cleanup(exchange.delete)

        return exchange

    return fabric


@pytest.fixture(autouse=True)
def memory_tracer():
    tracemalloc.start()
    tracemalloc.clear_traces()

    filters = (
        tracemalloc.Filter(True, aiormq.__file__),
        tracemalloc.Filter(True, pamqp.__file__),
        tracemalloc.Filter(True, aio_pika.__file__),
    )

    snapshot_before = tracemalloc.take_snapshot().filter_traces(filters)

    try:
        yield

        with suppress(Exception):
            gc.collect()

        snapshot_after = tracemalloc.take_snapshot().filter_traces(filters)

        top_stats = snapshot_after.compare_to(
            snapshot_before, "lineno", cumulative=True,
        )

        assert not top_stats
    finally:
        tracemalloc.stop()



import asyncio
import warnings
from collections import defaultdict
from itertools import chain
from typing import Any, DefaultDict, Dict, Optional, Set, Type, Union, cast
from warnings import warn

import aiormq

from .abc import (
    AbstractConnection, AbstractRobustChannel, AbstractRobustExchange,
    AbstractRobustQueue, TimeoutType,
)
from .channel import Channel
from .exchange import Exchange, ExchangeType
from .log import get_logger
from .queue import Queue
from .robust_exchange import RobustExchange
from .robust_queue import RobustQueue
from .tools import CallbackCollection


log = get_logger(__name__)


class RobustChannel(Channel, AbstractRobustChannel):
    """ Channel abstraction """

    QUEUE_CLASS: Type[Queue] = RobustQueue
    EXCHANGE_CLASS: Type[Exchange] = RobustExchange

    RESTORE_RETRY_DELAY: int = 2

    _exchanges: DefaultDict[str, AbstractRobustExchange]
    _queues: DefaultDict[str, Set[RobustQueue]]
    default_exchange: RobustExchange

    def __init__(
        self,
        connection: AbstractConnection,
        channel_number: Optional[int] = None,
        publisher_confirms: bool = True,
        on_return_raises: bool = False,
    ):
        """

        :param connection: :class:`aio_pika.adapter.AsyncioConnection` instance
        :param loop:
            Event loop (:func:`asyncio.get_event_loop()` when :class:`None`)
        :param future_store: :class:`aio_pika.common.FutureStore` instance
        :param publisher_confirms:
            False if you don't need delivery confirmations
            (in pursuit of performance)
        """

        super().__init__(
            connection=connection,
            channel_number=channel_number,
            publisher_confirms=publisher_confirms,
            on_return_raises=on_return_raises,
        )

        self._exchanges = defaultdict()
        self._queues = defaultdict(set)
        self._prefetch_count: int = 0
        self._prefetch_size: int = 0
        self._global_qos: bool = False
        self.reopen_callbacks = CallbackCollection(self)
        self.__restore_lock = asyncio.Lock()
        self.__restored = asyncio.Event()

        self.close_callbacks.remove(self._set_closed_callback)

    async def ready(self) -> None:
        await self._connection.ready()
        await self.__restored.wait()

    async def get_underlay_channel(self) -> aiormq.abc.AbstractChannel:
        await self._connection.ready()
        return await super().get_underlay_channel()

    async def restore(self, channel: Any = None) -> None:
        if channel is not None:
            warnings.warn(
                "Channel argument will be ignored because you "
                "don't need to pass this anymore.",
                DeprecationWarning,
            )

        async with self.__restore_lock:
            if self.__restored.is_set():
                return

            await self.reopen()
            self.__restored.set()

    async def _on_close(
        self,
        closing: asyncio.Future
    ) -> Optional[BaseException]:
        exc = await super()._on_close(closing)

        if isinstance(exc, asyncio.CancelledError):
            # Connection forced close (e.g., stuck connection).
            # Clear restored but do NOT set _closed - the robust connection
            # will restore this channel via restore() -> reopen() -> _open()
            # which properly handles _closed state reset.
            self.__restored.clear()
            return exc

        in_restore_state = not self.__restored.is_set()
        self.__restored.clear()

        if self._closed.done() or in_restore_state:
            return exc

        await self.restore()

        return exc

    async def close(
        self,
        exc: Optional[aiormq.abc.ExceptionType] = None,
    ) -> None:
        # Avoid recovery when channel is explicitely closed using this method
        self.__restored.clear()
        await super().close(exc)

    async def reopen(self) -> None:
        await super().reopen()
        await self.reopen_callbacks()

    async def _on_open(self) -> None:
        if not hasattr(self, "default_exchange"):
            await super()._on_open()

        exchanges = self._exchanges.values()
        queues = tuple(chain(*self._queues.values()))
        channel = await self.get_underlay_channel()

        await channel.basic_qos(
            prefetch_count=self._prefetch_count,
            prefetch_size=self._prefetch_size,
            global_=self._global_qos,
        )

        for exchange in exchanges:
            await exchange.restore()

        for queue in queues:
            await queue.restore()

        if hasattr(self, "default_exchange"):
            self.default_exchange.channel = self

        self.__restored.set()

    async def set_qos(
        self,
        prefetch_count: int = 0,
        prefetch_size: int = 0,
        global_: bool = False,
        timeout: TimeoutType = None,
        all_channels: Optional[bool] = None,
    ) -> aiormq.spec.Basic.QosOk:
        if all_channels is not None:
            warn('Use "global_" instead of "all_channels"', DeprecationWarning)
            global_ = all_channels

        await self.ready()

        self._prefetch_count = prefetch_count
        self._prefetch_size = prefetch_size
        self._global_qos = global_

        return await super().set_qos(
            prefetch_count=prefetch_count,
            prefetch_size=prefetch_size,
            global_=global_,
            timeout=timeout,
        )

    async def declare_exchange(
        self,
        name: str,
        type: Union[ExchangeType, str] = ExchangeType.DIRECT,
        durable: bool = False,
        auto_delete: bool = False,
        internal: bool = False,
        passive: bool = False,
        arguments: Optional[Dict[str, Any]] = None,
        timeout: TimeoutType = None,
        robust: bool = True,
    ) -> AbstractRobustExchange:
        """
        :param robust: If True, the exchange will be re-declared during
        reconnection.
        Set to False for temporary exchanges that should not be restored.
        """
        await self.ready()
        # Passive is True so expecting the exchange to be already declared
        # if we can just return it instead of creating a new class instance
        if passive and name in self._exchanges:
            return self._exchanges[name]

        exchange = (
            await super().declare_exchange(
                name=name,
                type=type,
                durable=durable,
                auto_delete=auto_delete,
                internal=internal,
                passive=passive,
                arguments=arguments,
                timeout=timeout,
            )
        )
        exchange = cast(AbstractRobustExchange, exchange)

        if not internal and robust:
            self._exchanges[name] = exchange

        return exchange

    async def exchange_delete(
        self,
        exchange_name: str,
        timeout: TimeoutType = None,
        if_unused: bool = False,
        nowait: bool = False,
    ) -> aiormq.spec.Exchange.DeleteOk:
        await self.ready()
        result = await super().exchange_delete(
            exchange_name=exchange_name,
            timeout=timeout,
            if_unused=if_unused,
            nowait=nowait,
        )
        self._exchanges.pop(exchange_name, None)
        return result

    async def declare_queue(
        self,
        name: Optional[str] = None,
        *,
        durable: bool = False,
        exclusive: bool = False,
        passive: bool = False,
        auto_delete: bool = False,
        arguments: Optional[Dict[str, Any]] = None,
        timeout: TimeoutType = None,
        robust: bool = True,
    ) -> AbstractRobustQueue:
        """
        :param robust: If True, the queue will be re-declared during
        reconnection.
        Set to False for temporary queues that should not be restored.
        """
        await self.ready()
        # Passive is True so expecting the queue to be already declared
        # if we can just return it instead of creating a new class instance
        if passive and name and name in self._queues:
            return list(self._queues[name])[0]

        queue = await super().declare_queue(
            name=name,
            durable=durable,
            exclusive=exclusive,
            passive=passive,
            auto_delete=auto_delete,
            arguments=arguments,
            timeout=timeout,
        )
        queue = cast(RobustQueue, queue)

        if robust:
            self._queues[queue.name].add(queue)
        return queue

    async def queue_delete(
        self,
        queue_name: str,
        timeout: TimeoutType = None,
        if_unused: bool = False,
        if_empty: bool = False,
        nowait: bool = False,
    ) -> aiormq.spec.Queue.DeleteOk:
        await self.ready()
        result = await super().queue_delete(
            queue_name=queue_name,
            timeout=timeout,
            if_unused=if_unused,
            if_empty=if_empty,
            nowait=nowait,
        )
        self._queues.pop(queue_name, None)
        return result


__all__ = ("RobustChannel",)



import aio_pika
import aio_pika.abc


async def test_connect_robust(amqp_url) -> None:
    async with await aio_pika.connect_robust(amqp_url) as connection:
        assert isinstance(connection, aio_pika.abc.AbstractRobustConnection)
        assert isinstance(connection, aio_pika.abc.AbstractConnection)

        channel = await connection.channel()
        assert isinstance(channel, aio_pika.abc.AbstractRobustChannel)
        assert isinstance(channel, aio_pika.abc.AbstractChannel)


async def test_connect(amqp_url) -> None:
    async with await aio_pika.connect(amqp_url) as connection:
        assert isinstance(connection, aio_pika.abc.AbstractConnection)
        assert not isinstance(
            connection, aio_pika.abc.AbstractRobustConnection,
        )

        channel = await connection.channel()
        assert isinstance(channel, aio_pika.abc.AbstractChannel)
        assert not isinstance(
            channel, aio_pika.abc.AbstractRobustChannel,
        )



import gc
import weakref
from functools import partial

import pytest

import aio_pika
from aio_pika import RobustChannel
from tests import get_random_name


async def test_leak_unclosed_channel(create_connection):
    rabbitmq_connection = await create_connection()

    weakset: weakref.WeakSet[aio_pika.abc.AbstractChannel] = weakref.WeakSet()

    async def f(rabbitmq_connection: aio_pika.Connection, weakset):
        weakset.add(await rabbitmq_connection.channel())

    async with rabbitmq_connection:
        for i in range(5):
            await f(rabbitmq_connection, weakset)

    gc.collect()

    assert len(tuple(weakset)) == 0


async def test_leak_closed_channel(create_connection):
    rabbitmq_connection = await create_connection()

    weakset: weakref.WeakSet[aio_pika.abc.AbstractConnection] = (
        weakref.WeakSet()
    )

    async def f(rabbitmq_connection: aio_pika.Connection, weakset):
        async with rabbitmq_connection.channel() as channel:
            weakset.add(channel)

    async with rabbitmq_connection:
        for i in range(5):
            await f(rabbitmq_connection, weakset)

    gc.collect()

    assert len(tuple(weakset)) == 0


@pytest.fixture
def robust_connection_fabric():
    return aio_pika.connect_robust


@pytest.fixture
def create_robust_connection(robust_connection_fabric, event_loop, amqp_url):
    return partial(robust_connection_fabric, amqp_url, loop=event_loop)


async def test_get_exchange_no_memory_leak(create_robust_connection):
    """Test that repeated get_exchange calls don't create multiple instances.

    This verifies the fix for the memory leak where each call to get_exchange()
    was adding a new exchange instance to the internal storage instead of
    returning the cached one.
    """
    connection = await create_robust_connection()

    weakset: weakref.WeakSet[aio_pika.abc.AbstractExchange] = weakref.WeakSet()

    async with connection:
        channel: RobustChannel = await connection.channel()  # type: ignore
        name = get_random_name("memory", "leak", "exchange")

        # Declare the exchange
        exchange = await channel.declare_exchange(
            name, auto_delete=True,
        )
        weakset.add(exchange)

        # Get the exchange multiple times - should return same instance
        for _ in range(10):
            ex = await channel.get_exchange(name)
            weakset.add(ex)

        # Should only have 1 unique instance in the WeakSet
        assert len(tuple(weakset)) == 1

        # Verify internal storage doesn't grow
        assert name in channel._exchanges
        # Should be a single exchange, not a set/list of exchanges
        assert isinstance(
            channel._exchanges[name],
            aio_pika.abc.AbstractExchange
        )


async def test_get_queue_no_memory_leak(create_robust_connection):
    """Test that repeated get_queue calls don't create multiple instances.

    This verifies the fix for the memory leak where each call to get_queue()
    with passive=True was creating new queue instances instead of returning
    the cached one.
    """
    connection = await create_robust_connection()

    weakset: weakref.WeakSet[aio_pika.abc.AbstractQueue] = weakref.WeakSet()

    async with connection:
        channel: RobustChannel = await connection.channel()  # type: ignore
        name = get_random_name("memory", "leak", "queue")

        # Declare the queue
        queue = await channel.declare_queue(
            name, auto_delete=True,
        )
        weakset.add(queue)

        # Get the queue multiple times - should return same instance
        for _ in range(10):
            q = await channel.get_queue(name)
            weakset.add(q)

        # Should only have 1 unique instance in the WeakSet
        assert len(tuple(weakset)) == 1

        # Verify internal storage has only one queue for this name
        assert name in channel._queues
        assert len(channel._queues[name]) == 1



import asyncio
from functools import partial

import pytest
from aiormq import ChannelNotFoundEntity
from aiormq.exceptions import ChannelPreconditionFailed

import aio_pika
from aio_pika import RobustChannel
from tests import get_random_name
from tests.test_amqp import (
    TestCaseAmqp, TestCaseAmqpNoConfirms, TestCaseAmqpWithConfirms,
)


@pytest.fixture
def connection_fabric():
    return aio_pika.connect_robust


@pytest.fixture
def create_connection(connection_fabric, event_loop, amqp_url):
    return partial(connection_fabric, amqp_url, loop=event_loop)


class TestCaseNoRobust(TestCaseAmqp):   # type: ignore
    PARAMS = [{"robust": True}, {"robust": False}]
    IDS = ["robust=1", "robust=0"]

    @staticmethod
    @pytest.fixture(name="declare_queue", params=PARAMS, ids=IDS)
    def declare_queue_(request, declare_queue):     # type: ignore
        async def fabric(*args, **kwargs) -> aio_pika.Queue:
            kwargs.update(request.param)
            return await declare_queue(*args, **kwargs)

        return fabric

    @staticmethod
    @pytest.fixture(name="declare_exchange", params=PARAMS, ids=IDS)
    def declare_exchange_(request, declare_exchange):   # type: ignore
        async def fabric(*args, **kwargs) -> aio_pika.Queue:
            kwargs.update(request.param)
            return await declare_exchange(*args, **kwargs)

        return fabric

    async def test_add_reconnect_callback(self, create_connection):
        connection = await create_connection()

        def cb(*a, **kw):
            pass

        connection.reconnect_callbacks.add(cb)

        del cb
        assert len(connection.reconnect_callbacks) == 1

    async def test_channel_blocking_timeout_reopen(self, connection):
        channel: RobustChannel = await connection.channel()     # type: ignore
        close_reasons = []
        close_event = asyncio.Event()
        reopen_event = asyncio.Event()
        channel.reopen_callbacks.add(lambda *_: reopen_event.set())

        queue_name = get_random_name("test_channel_blocking_timeout_reopen")

        def on_done(*args):
            close_reasons.append(args)
            close_event.set()
            return

        channel.close_callbacks.add(on_done)

        with pytest.raises(ChannelNotFoundEntity):
            await channel.declare_queue(queue_name, passive=True)

        await close_event.wait()
        assert channel.is_closed

        # Ensure close callback has been called
        assert close_reasons

        await asyncio.wait_for(reopen_event.wait(), timeout=60)
        await channel.declare_queue(queue_name, auto_delete=True)

    async def test_get_queue_fail(self, connection):
        channel: RobustChannel = await connection.channel()     # type: ignore
        close_event = asyncio.Event()
        reopen_event = asyncio.Event()
        channel.close_callbacks.add(lambda *_: close_event.set())
        channel.reopen_callbacks.add(lambda *_: reopen_event.set())

        name = get_random_name("passive", "queue")

        await channel.declare_queue(
            name,
            auto_delete=True,
            arguments={"x-max-length": 1},
        )
        with pytest.raises(ChannelPreconditionFailed):
            await channel.declare_queue(name, auto_delete=True)
        await asyncio.sleep(0)
        await close_event.wait()
        await reopen_event.wait()
        with pytest.raises(ChannelPreconditionFailed):
            await channel.declare_queue(name, auto_delete=True)

    async def test_channel_is_ready_after_close_and_reopen(self, connection):
        channel: RobustChannel = await connection.channel()  # type: ignore
        await channel.ready()
        await channel.close()
        assert channel.is_closed is True

        await channel.reopen()
        await asyncio.wait_for(channel.ready(), timeout=1)

        assert channel.is_closed is False

    async def test_channel_can_be_closed(self, connection):
        channel: RobustChannel = await connection.channel()  # type: ignore
        await channel.ready()
        await channel.close()

        assert channel.is_closed

        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(channel.ready(), timeout=1)

        assert channel.is_closed

    async def test_get_exchange(self, connection, declare_exchange):
        channel = await self.create_channel(connection)
        name = get_random_name("passive", "exchange")

        with pytest.raises(aio_pika.exceptions.ChannelNotFoundEntity):
            await channel.get_exchange(name)

        channel = await self.create_channel(connection)
        exchange = await declare_exchange(
            name, auto_delete=True, channel=channel,
        )
        exchange_passive = await channel.get_exchange(name)

        assert exchange.name is exchange_passive.name

    async def test_get_exchange_returns_same_instance(self, connection):
        """Test that get_exchange returns the cached instance on repeated calls.

        This tests the fix for the memory leak where each call to get_exchange
        (which uses passive=True) was creating a new exchange instance instead
        of returning the cached one.

        Note: This only applies when robust=True (default), as non-robust
        exchanges are not cached.
        """
        channel = await self.create_channel(connection)
        name = get_random_name("passive", "exchange", "memory")

        # Declare exchange with robust=True (default) so it gets cached
        exchange = await channel.declare_exchange(
            name, auto_delete=True,
        )

        # Get the exchange multiple times - should return the same instance
        exchange_passive1 = await channel.get_exchange(name)
        exchange_passive2 = await channel.get_exchange(name)
        exchange_passive3 = await channel.get_exchange(name)

        # All should be the exact same object (same id)
        assert exchange is exchange_passive1
        assert exchange_passive1 is exchange_passive2
        assert exchange_passive2 is exchange_passive3

    async def test_get_queue_returns_same_instance(self, connection):
        """Test that get_queue returns the cached instance on repeated calls.

        This tests the fix for the memory leak where each call to get_queue
        (which uses passive=True) was creating a new queue instance instead
        of returning the cached one.

        Note: This only applies when robust=True (default), as non-robust
        queues are not cached.
        """
        channel = await self.create_channel(connection)
        name = get_random_name("passive", "queue", "memory")

        # Declare queue with robust=True (default) so it gets cached
        queue = await channel.declare_queue(
            name, auto_delete=True,
        )

        # Get the queue multiple times using passive=True - should return same
        queue_passive1 = await channel.get_queue(name)
        queue_passive2 = await channel.get_queue(name)
        queue_passive3 = await channel.get_queue(name)

        # All should be the exact same object (same id)
        assert queue is queue_passive1
        assert queue_passive1 is queue_passive2
        assert queue_passive2 is queue_passive3


class TestCaseAmqpNoConfirmsRobust(TestCaseAmqpNoConfirms):
    pass


class TestCaseAmqpWithConfirmsRobust(TestCaseAmqpWithConfirms):
    pass



import asyncio
import logging
import os
import time
import uuid
from datetime import datetime, timezone
from typing import Callable, Optional, List
from unittest import mock

import aiormq.exceptions
import pytest
import shortuuid
from yarl import URL

import aio_pika
import aio_pika.exceptions
from aio_pika import Channel, DeliveryMode, Message
from aio_pika.abc import (
    AbstractConnection, AbstractIncomingMessage, MessageInfo, AbstractQueue,
)
from aio_pika.exceptions import (
    DeliveryError, MessageProcessError, ProbableAuthenticationError,
)
from aio_pika.exchange import ExchangeType
from aio_pika.message import ReturnedMessage
from aio_pika.queue import QueueIterator
from tests import get_random_name


log = logging.getLogger(__name__)


class TestCaseAmqpBase:
    @staticmethod
    def create_channel(
        connection: aio_pika.Connection,
    ) -> aio_pika.abc.AbstractChannel:
        return connection.channel()

    @staticmethod
    @pytest.fixture(name="declare_queue")
    def declare_queue_(declare_queue):
        return declare_queue

    @staticmethod
    @pytest.fixture(name="declare_exchange")
    def declare_exchange_(declare_exchange):
        return declare_exchange


class TestCaseAmqp(TestCaseAmqpBase):
    async def test_properties(
        self, event_loop, connection: aio_pika.Connection,
    ):
        assert not connection.is_closed

    async def test_channel_close(self, connection: aio_pika.Connection):
        event = asyncio.Event()

        closed = False

        def on_close(
            ch: Optional[aio_pika.abc.AbstractChannel],
            exc: Optional[BaseException] = None,
        ):
            nonlocal closed
            log.info("Close called")
            closed = True
            assert ch is not None
            assert ch.is_closed
            event.set()

        channel = await self.create_channel(connection)
        channel.close_callbacks.add(on_close)
        await channel.close()

        await event.wait()

        assert closed

        with pytest.raises(RuntimeError):
            await channel.initialize()

        async with self.create_channel(connection) as ch:
            assert not ch.is_closed

    async def test_channel_reopen(self, connection):
        channel = await self.create_channel(connection)

        await channel.close()
        assert channel.is_closed

        await channel.reopen()
        assert not channel.is_closed

    async def test_delete_queue_and_exchange(
        self, connection, declare_exchange, declare_queue,
    ):
        queue_name = get_random_name("test_connection")
        exchange = get_random_name()

        channel = await self.create_channel(connection)
        await declare_exchange(exchange, auto_delete=True)
        await declare_queue(queue_name, auto_delete=True)

        await channel.queue_delete(queue_name)
        await channel.exchange_delete(exchange)

    async def test_temporary_queue(self, connection, declare_queue):
        channel = await self.create_channel(connection)
        queue = await declare_queue(auto_delete=True)

        assert queue.name != ""

        body = os.urandom(32)

        await channel.default_exchange.publish(
            Message(body=body), routing_key=queue.name,
        )

        await asyncio.sleep(1)

        message = await queue.get()

        assert message.body == body

        await channel.queue_delete(queue.name)

    @pytest.mark.skip(reason="This was deprecated in AMQP 0-9-1")
    async def test_internal_exchange(
        self, channel: aio_pika.Channel, declare_exchange, declare_queue,
    ):
        routing_key = get_random_name()
        exchange_name = get_random_name("internal", "exchange")

        exchange = await declare_exchange(
            exchange_name, auto_delete=True, internal=True,
        )

        queue = await declare_queue(auto_delete=True)

        await queue.bind(exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        with pytest.raises(ValueError):
            f = exchange.publish(
                Message(
                    body, content_type="text/plain", headers={"foo": "bar"},
                ),
                routing_key,
            )
            await f

        await queue.unbind(exchange, routing_key)

    async def test_declare_exchange_with_passive_flag(
        self, connection, declare_exchange: Callable,
    ):
        exchange_name = get_random_name()
        channel = await self.create_channel(connection)

        with pytest.raises(aio_pika.exceptions.ChannelNotFoundEntity):
            await declare_exchange(
                exchange_name, auto_delete=True, passive=True, channel=channel,
            )

        channel1 = await self.create_channel(connection)
        channel2 = await self.create_channel(connection)

        await declare_exchange(
            exchange_name, auto_delete=True, passive=False, channel=channel1,
        )

        # Check ignoring different exchange options
        await declare_exchange(
            exchange_name, auto_delete=False, passive=True, channel=channel2,
        )

    async def test_declare_queue_with_passive_flag(
        self, connection, declare_exchange: Callable, declare_queue: Callable,
    ):
        queue_name = get_random_name()
        ch1 = await self.create_channel(connection)
        ch2 = await self.create_channel(connection)
        ch3 = await self.create_channel(connection)

        with pytest.raises(aio_pika.exceptions.ChannelNotFoundEntity):
            await declare_queue(
                queue_name, auto_delete=True, passive=True, channel=ch1,
            )

        await declare_queue(
            queue_name, auto_delete=True, passive=False, channel=ch2,
        )

        # Check ignoring different queue options
        await declare_queue(
            queue_name, auto_delete=False, passive=True, channel=ch3,
        )

    async def test_simple_publish_and_receive(
        self,
        channel: aio_pika.Channel,
        declare_queue: Callable,
        declare_exchange: Callable,
    ):
        queue_name = get_random_name("test_connection")
        routing_key = get_random_name()

        exchange = await declare_exchange(
            "direct", auto_delete=True, channel=channel,
        )

        queue = await declare_queue(
            queue_name, auto_delete=True, channel=channel,
        )

        await queue.bind(exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        result = await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        assert result

        incoming_message = await queue.get(timeout=5)
        await incoming_message.ack()

        assert incoming_message.body == body

        await queue.unbind(exchange, routing_key)

    async def test_simple_publish_without_confirm(
        self,
        connection: aio_pika.Connection,
        declare_exchange: Callable,
        declare_queue: Callable,
    ):
        queue_name = get_random_name("test_connection")
        routing_key = get_random_name()

        channel = await connection.channel(publisher_confirms=False)

        exchange = await declare_exchange(
            "direct", auto_delete=True, channel=channel,
        )
        queue = await declare_queue(
            queue_name, auto_delete=True, channel=channel,
        )

        await queue.bind(exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        result = await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        assert result is None

        incoming_message = await queue.get(timeout=5)
        await incoming_message.ack()

        assert incoming_message.body == body

        await queue.unbind(exchange, routing_key)

    async def test_simple_publish_and_receive_delivery_mode_explicitly(
        self,
        channel: aio_pika.Channel,
        declare_queue: Callable,
        declare_exchange: Callable,
    ):
        queue_name = get_random_name("test_connection")
        routing_key = get_random_name()

        exchange = await declare_exchange(
            "direct", auto_delete=True, channel=channel,
        )
        queue = await declare_queue(
            queue_name, auto_delete=True, channel=channel,
        )

        await queue.bind(exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        await exchange.publish(
            Message(
                body,
                content_type="text/plain",
                headers={"foo": "bar"},
                delivery_mode=None,
            ),
            routing_key,
        )

        incoming_message = await queue.get(timeout=5)
        await incoming_message.ack()

        assert incoming_message.body == body

        await queue.unbind(exchange, routing_key)

    async def test_simple_publish_and_receive_to_bound_exchange(
        self,
        channel: aio_pika.Channel,
        declare_exchange: Callable,
        declare_queue: Callable,
        add_cleanup: Callable,
    ):
        routing_key = get_random_name()
        src_name = get_random_name("source", "exchange")
        dest_name = get_random_name("destination", "exchange")

        src_exchange = await declare_exchange(src_name, auto_delete=True)
        dest_exchange = await declare_exchange(dest_name, auto_delete=True)
        queue = await declare_queue(auto_delete=True)

        await queue.bind(dest_exchange, routing_key)

        await dest_exchange.bind(src_exchange, routing_key)
        add_cleanup(dest_exchange.unbind, src_exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        await src_exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        incoming_message = await queue.get(timeout=5)
        await incoming_message.ack()

        assert incoming_message.body == body

        await queue.unbind(dest_exchange, routing_key)

    async def test_simple_publish_with_closed_channel(
        self,
        connection: aio_pika.Connection,
        declare_exchange: Callable,
        declare_queue: Callable,
    ):
        routing_key = get_random_name()

        channel = await connection.channel(publisher_confirms=False)

        exchange = await declare_exchange(
            "direct", auto_delete=True, channel=channel,
        )

        await connection.close()

        body = bytes(shortuuid.uuid(), "utf-8")

        with pytest.raises(aiormq.exceptions.ChannelInvalidStateError):
            await exchange.publish(
                Message(
                    body,
                    content_type="text/plain",
                    headers={"foo": "bar"},
                ),
                routing_key,
            )

    async def test_incoming_message_info(
        self,
        channel: aio_pika.Channel,
        declare_queue: Callable,
        declare_exchange: Callable,
        add_cleanup: Callable,
    ):
        queue_name = get_random_name("test_connection")
        routing_key = get_random_name()

        exchange = await declare_exchange("direct", auto_delete=True)
        queue = await declare_queue(queue_name, auto_delete=True)

        await queue.bind(exchange, routing_key)
        add_cleanup(queue.unbind, exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        info = MessageInfo(
            app_id="test",
            body_size=len(body),
            content_encoding="text",
            content_type="application/json",
            correlation_id="1",
            delivery_mode=DeliveryMode.PERSISTENT,
            expiration=1.5,
            headers={"foo": "bar"},
            message_id=shortuuid.uuid(),
            priority=0,
            reply_to="test",
            timestamp=datetime.fromtimestamp(int(time.time()), tz=timezone.utc),
            type="0",
            user_id="guest",
        )

        msg = Message(
            body=body,
            headers={"foo": "bar"},
            content_type="application/json",
            content_encoding="text",
            delivery_mode=DeliveryMode.PERSISTENT,
            priority=0,
            correlation_id="1",
            reply_to="test",
            expiration=1.5,
            message_id=info["message_id"],
            timestamp=info["timestamp"],
            type="0",
            user_id="guest",
            app_id="test",
        )

        await exchange.publish(msg, routing_key)

        incoming_message = await queue.get(timeout=5)
        await incoming_message.ack()

        info["routing_key"] = incoming_message.routing_key
        info["redelivered"] = incoming_message.redelivered
        info["exchange"] = incoming_message.exchange
        info["delivery_tag"] = incoming_message.delivery_tag
        info["consumer_tag"] = incoming_message.consumer_tag
        info["cluster_id"] = incoming_message.cluster_id

        assert incoming_message.body == body
        assert incoming_message.info() == info

    async def test_context_process(
        self,
        channel: aio_pika.Channel,
        declare_queue: Callable,
        declare_exchange: Callable,
        add_cleanup: Callable,
    ):
        queue_name = get_random_name("test_connection")
        routing_key = get_random_name()

        exchange = await declare_exchange("direct", auto_delete=True)
        queue = await declare_queue(queue_name, auto_delete=True)

        await queue.bind(exchange, routing_key)
        add_cleanup(queue.unbind, exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        if not channel.publisher_confirms:
            await asyncio.sleep(1)

        incoming_message: AbstractIncomingMessage = await queue.get(timeout=5)

        with pytest.raises(AssertionError):
            async with incoming_message.process(requeue=True):
                raise AssertionError

        assert incoming_message.locked

        incoming_message = await queue.get(timeout=5)

        async with incoming_message.process():
            pass

        assert incoming_message.body == body

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        incoming_message = await queue.get(timeout=5)

        with pytest.raises(MessageProcessError):
            async with incoming_message.process():
                await incoming_message.reject(requeue=True)

        assert incoming_message.locked

        incoming_message = await queue.get(timeout=5)

        async with incoming_message.process(ignore_processed=True):
            await incoming_message.reject(requeue=False)

        assert incoming_message.body == body

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        incoming_message = await queue.get(timeout=5)
        with pytest.raises(AssertionError):
            async with incoming_message.process(
                requeue=True, reject_on_redelivered=True,
            ):
                raise AssertionError

        incoming_message = await queue.get(timeout=5)
        with pytest.raises(AssertionError):
            async with incoming_message.process(
                requeue=True, reject_on_redelivered=True,
            ):
                raise AssertionError

        assert incoming_message.locked

    async def test_context_process_redelivery(
        self,
        channel: aio_pika.Channel,
        declare_exchange: Callable,
        declare_queue: Callable,
        add_cleanup: Callable,
    ):
        queue_name = get_random_name("test_connection")
        routing_key = get_random_name()

        exchange = await declare_exchange("direct", auto_delete=True)
        queue = await declare_queue(queue_name, auto_delete=True)

        await queue.bind(exchange, routing_key)
        add_cleanup(queue.unbind, exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        if not channel.publisher_confirms:
            await asyncio.sleep(1)

        incoming_message = await queue.get(timeout=5)

        with pytest.raises(AssertionError):
            async with incoming_message.process(
                requeue=True, reject_on_redelivered=True,
            ):
                raise AssertionError

        incoming_message = await queue.get(timeout=5)

        with mock.patch("aio_pika.message.log") as message_logger:
            with pytest.raises(Exception):
                async with incoming_message.process(
                    requeue=True, reject_on_redelivered=True,
                ):
                    raise Exception

            assert message_logger.info.called
            assert (
                message_logger.info.mock_calls[0][1][1].body
                == incoming_message.body
            )

        assert incoming_message.body == body

    async def test_no_ack_redelivery(
        self,
        connection,
        add_cleanup: Callable,
        declare_queue,
        declare_exchange,
    ):
        queue_name = get_random_name("test_connection")
        routing_key = get_random_name()

        channel = await self.create_channel(connection)
        exchange = await declare_exchange(
            "direct", auto_delete=True, channel=channel,
        )
        queue = await declare_queue(
            queue_name, auto_delete=False, channel=channel, cleanup=False,
        )

        await queue.bind(exchange, routing_key)

        # publish 2 messages
        for _ in range(2):
            body = bytes(shortuuid.uuid(), "utf-8")
            msg = Message(body)
            await exchange.publish(msg, routing_key)

        if not channel.publisher_confirms:
            await asyncio.sleep(1)

        # ack 1 message out of 2
        first_message = await queue.get(timeout=5)

        last_message = await queue.get(timeout=5)
        await last_message.ack()

        # close channel, not acked message should be redelivered
        await channel.close()

        channel = await self.create_channel(connection)
        exchange = await declare_exchange(
            "direct", auto_delete=True, channel=channel,
        )
        queue = await declare_queue(
            queue_name, auto_delete=False, channel=channel,
        )

        # receive not acked message
        message = await queue.get(timeout=5)
        assert message.body == first_message.body
        await message.ack()

        await queue.unbind(exchange, routing_key)

    async def test_ack_multiple(
        self,
        connection,
        declare_exchange,
        declare_queue,
        add_cleanup: Callable,
    ):
        queue_name = get_random_name("test_connection")
        routing_key = get_random_name()

        channel = await self.create_channel(connection)
        exchange = await declare_exchange(
            "direct", auto_delete=True, channel=channel,
        )
        queue = await declare_queue(
            queue_name, auto_delete=False, cleanup=False, channel=channel,
        )

        await queue.bind(exchange, routing_key)

        # publish 2 messages
        for _ in range(2):
            body = bytes(shortuuid.uuid(), "utf-8")
            msg = Message(body)
            await exchange.publish(msg, routing_key)

        if not channel.publisher_confirms:
            await asyncio.sleep(1)

        # ack only last mesage with multiple flag, first
        # message should be acked too
        await queue.get(timeout=5)
        last_message = await queue.get(timeout=5)
        await last_message.ack(multiple=True)

        # close channel, no messages should be redelivered
        await channel.close()

        channel = await self.create_channel(connection)
        exchange = await declare_exchange(
            "direct", auto_delete=True, channel=channel,
        )
        queue = await declare_queue(
            queue_name, auto_delete=False, cleanup=False, channel=channel,
        )

        with pytest.raises(aio_pika.exceptions.QueueEmpty):
            await queue.get()

        await queue.unbind(exchange, routing_key)
        await queue.delete()

    async def test_ack_twice(
        self, channel: aio_pika.Connection, declare_queue, declare_exchange,
    ):
        queue_name = get_random_name("test_connection")
        routing_key = get_random_name()

        exchange = await declare_exchange("direct", auto_delete=True)
        queue = await declare_queue(queue_name, auto_delete=True)

        await queue.bind(exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        incoming_message = await queue.get(timeout=5)
        await incoming_message.ack()

        with pytest.raises(MessageProcessError):
            await incoming_message.ack()

        assert incoming_message.body == body

        await queue.unbind(exchange, routing_key)
        await queue.delete()

    async def test_reject_twice(
        self,
        channel: aio_pika.Channel,
        add_cleanup: Callable,
        declare_queue: Callable,
        declare_exchange: Callable,
    ):
        queue_name = get_random_name("test_connection")
        routing_key = get_random_name()

        exchange = await declare_exchange("direct", auto_delete=True)
        queue = await declare_queue(queue_name, auto_delete=True)

        await queue.bind(exchange, routing_key)
        add_cleanup(queue.unbind, exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        incoming_message = await queue.get(timeout=5)
        await incoming_message.reject(requeue=False)

        with pytest.raises(MessageProcessError):
            await incoming_message.reject(requeue=False)

        assert incoming_message.body == body

    async def test_consuming(
        self,
        event_loop,
        channel: aio_pika.Channel,
        declare_exchange: Callable,
        declare_queue: Callable,
        add_cleanup: Callable,
    ):
        queue_name = get_random_name("tc2")
        routing_key = get_random_name()

        exchange = await declare_exchange("direct", auto_delete=True)
        queue = await declare_queue(queue_name, auto_delete=True)

        await queue.bind(exchange, routing_key)
        add_cleanup(queue.unbind, exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        f = event_loop.create_future()

        async def handle(message):
            await message.ack()
            assert message.body == body
            assert message.routing_key == routing_key
            f.set_result(True)

        await queue.consume(handle)

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        if not f.done():
            await f

    async def test_consuming_not_coroutine(
        self,
        event_loop,
        channel: aio_pika.Channel,
        declare_exchange: Callable,
        declare_queue: Callable,
        add_cleanup: Callable,
    ):

        queue_name = get_random_name("tc2")
        routing_key = get_random_name()

        exchange = await declare_exchange("direct", auto_delete=True)
        queue = await declare_queue(queue_name, auto_delete=True)
        add_cleanup(queue.unbind, exchange, routing_key)

        await queue.bind(exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        f = event_loop.create_future()

        async def handle(message):
            await message.ack()
            assert message.body == body
            assert message.routing_key == routing_key
            f.set_result(True)

        await queue.consume(handle)

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        if not f.done():
            await f

    async def test_ack_reject(
        self,
        channel: aio_pika.Channel,
        declare_exchange: Callable,
        declare_queue: Callable,
        add_cleanup: Callable,
    ):
        queue_name = get_random_name("test_connection3")
        routing_key = get_random_name()

        exchange = await declare_exchange("direct", auto_delete=True)
        queue = await declare_queue(queue_name, auto_delete=True)

        await queue.bind(exchange, routing_key)
        add_cleanup(queue.unbind, exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        incoming_message = await queue.get(timeout=5, no_ack=True)

        with pytest.raises(TypeError):
            await incoming_message.ack()

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        incoming_message = await queue.get(timeout=5)

        await incoming_message.reject()

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        incoming_message = await queue.get(timeout=5, no_ack=True)

        with pytest.raises(TypeError):
            await incoming_message.reject()

        assert incoming_message.body == body

    async def test_purge_queue(
        self,
        declare_queue: Callable,
        declare_exchange: Callable,
        channel: aio_pika.Channel,
    ):
        queue_name = get_random_name("test_connection4")
        routing_key = get_random_name()

        exchange = await declare_exchange("direct", auto_delete=True)
        queue = await declare_queue(queue_name, auto_delete=True)

        await queue.bind(exchange, routing_key)

        try:
            body = bytes(shortuuid.uuid(), "utf-8")

            await exchange.publish(
                Message(
                    body, content_type="text/plain", headers={"foo": "bar"},
                ),
                routing_key,
            )

            await queue.purge()

            with pytest.raises(asyncio.TimeoutError):
                await queue.get(timeout=1)
        except aio_pika.exceptions.QueueEmpty:
            await queue.unbind(exchange, routing_key)
            await queue.delete()

    async def test_connection_refused(self, connection_fabric: Callable):
        with pytest.raises(ConnectionError):
            await connection_fabric("amqp://guest:guest@localhost:9999")

    async def test_wrong_credentials(
        self, connection_fabric: Callable, amqp_url,
    ):
        amqp_url = amqp_url.with_user(uuid.uuid4().hex).with_password(
            uuid.uuid4().hex,
        )

        with pytest.raises(ProbableAuthenticationError):
            await connection_fabric(amqp_url)

    async def test_set_qos(self, channel: aio_pika.Channel):
        await channel.set_qos(prefetch_count=1, global_=True)

    async def test_set_qos_deprecated_all_channels(
        self, channel: aio_pika.Channel,
    ):
        with pytest.deprecated_call():
            await channel.set_qos(prefetch_count=1, all_channels=True)

    async def test_exchange_delete(
        self, channel: aio_pika.Channel, declare_exchange,
    ):
        exchange = await declare_exchange("test", auto_delete=True)
        await exchange.delete()

    async def test_dlx(
        self,
        event_loop,
        channel: aio_pika.Channel,
        declare_exchange: Callable,
        declare_queue: Callable,
        add_cleanup: Callable,
    ):
        suffix = get_random_name()
        routing_key = "%s_routing_key" % suffix
        dlx_routing_key = "%s_dlx_routing_key" % suffix

        f = event_loop.create_future()

        async def dlx_handle(message):
            await message.ack()
            assert message.body == body
            assert message.routing_key == dlx_routing_key
            f.set_result(True)

        direct_exchange = await declare_exchange(
            "direct", channel=channel, auto_delete=True,
        )  # type: aio_pika.Exchange

        dlx_exchange = await declare_exchange(
            "dlx", ExchangeType.DIRECT, auto_delete=True,
        )

        direct_queue = await declare_queue(
            "%s_direct_queue" % suffix,
            auto_delete=True,
            arguments={
                "x-message-ttl": 300,
                "x-dead-letter-exchange": "dlx",
                "x-dead-letter-routing-key": dlx_routing_key,
            },
        )

        dlx_queue = await declare_queue(
            "%s_dlx_queue" % suffix, auto_delete=True,
        )

        await dlx_queue.consume(dlx_handle)
        await dlx_queue.bind(dlx_exchange, dlx_routing_key)
        await direct_queue.bind(direct_exchange, routing_key)

        add_cleanup(dlx_queue.unbind, dlx_exchange, routing_key)
        add_cleanup(direct_queue.unbind, direct_exchange, routing_key)

        body = bytes(shortuuid.uuid(), "utf-8")

        await direct_exchange.publish(
            Message(
                body,
                content_type="text/plain",
                headers={
                    "x-message-ttl": 100,
                    "x-dead-letter-exchange": "dlx",
                },
            ),
            routing_key,
        )

        if not f.done():
            await f

    async def test_expiration(
        self, channel: aio_pika.Channel, event_loop,
        declare_exchange, declare_queue,
    ):

        dlx_queue = await declare_queue(
            get_random_name("test_dlx"), cleanup=False,
        )  # type: aio_pika.Queue

        dlx_exchange = await declare_exchange(
            get_random_name("dlx"), cleanup=False,
        )  # type: aio_pika.Exchange

        await dlx_queue.bind(dlx_exchange, routing_key=dlx_queue.name)

        queue = await declare_queue(
            get_random_name("test_expiration"),
            arguments={
                "x-message-ttl": 10000,
                "x-dead-letter-exchange": dlx_exchange.name,
                "x-dead-letter-routing-key": dlx_queue.name,
            },
        )  # type: aio_pika.Queue

        body = bytes(shortuuid.uuid(), "utf-8")

        await channel.default_exchange.publish(
            Message(
                body,
                content_type="text/plain",
                headers={"foo": "bar"},
                expiration=0.5,
            ),
            queue.name,
        )

        f = event_loop.create_future()

        await dlx_queue.consume(f.set_result, no_ack=True)

        message = await f

        assert message.body == body
        assert message.headers["x-death"][0]["original-expiration"] == "500"

    async def test_add_close_callback(self, create_connection: Callable):
        connection = await create_connection()

        shared_list = []

        def share(*a, **kw):
            shared_list.append((a, kw))

        connection.close_callbacks.add(share)

        del share
        assert len(connection.close_callbacks) == 1

        await connection.close()

        assert len(shared_list) == 1

    async def test_big_message(
        self,
        channel: aio_pika.Channel,
        add_cleanup: Callable,
        declare_queue,
        declare_exchange,
    ):
        queue_name = get_random_name("test_big")
        routing_key = get_random_name()

        exchange = await declare_exchange("direct", auto_delete=True)
        queue = await declare_queue(queue_name, auto_delete=True)

        await queue.bind(exchange, routing_key)
        add_cleanup(queue.unbind, exchange, routing_key)
        add_cleanup(queue.delete)

        body = bytes(shortuuid.uuid(), "utf-8") * 1000000

        await exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            routing_key,
        )

        incoming_message = await queue.get(timeout=5)
        await incoming_message.ack()

        assert incoming_message.body == body

    async def test_unexpected_channel_close(
        self, channel: aio_pika.Channel, declare_queue,
    ):
        with pytest.raises(aio_pika.exceptions.ChannelClosed):
            await declare_queue("amq.restricted_queue_name", auto_delete=True)

        with pytest.raises(aiormq.exceptions.ChannelInvalidStateError):
            await channel.set_qos(100)

    async def test_declaration_result(
        self, channel: aio_pika.Channel, declare_queue,
    ):
        queue = await declare_queue(auto_delete=True)
        assert queue.declaration_result.message_count == 0
        assert queue.declaration_result.consumer_count == 0

    async def test_declaration_result_with_consumers(
        self, connection, declare_queue,
    ):
        channel1 = await self.create_channel(connection)
        channel2 = await self.create_channel(connection)

        queue_name = get_random_name("queue", "declaration-result")
        queue1 = await declare_queue(
            queue_name, auto_delete=True, channel=channel1,
        )
        await queue1.consume(print)

        queue2 = await declare_queue(
            queue_name, passive=True, channel=channel2, cleanup=False,
        )

        assert queue2.declaration_result.consumer_count == 1

    async def test_declaration_result_with_messages(
        self, connection, declare_queue, declare_exchange,
    ):
        channel1 = await self.create_channel(connection)
        channel2 = await self.create_channel(connection)

        queue_name = get_random_name("queue", "declaration-result")
        queue1 = await declare_queue(
            queue_name, auto_delete=True, channel=channel1,
        )

        await channel1.default_exchange.publish(
            Message(body=b"test"), routing_key=queue1.name,
        )

        await asyncio.sleep(1)

        queue2 = await declare_queue(
            queue_name, passive=True, channel=channel2,
        )
        await queue2.get()
        await queue2.delete()

        assert queue2.declaration_result.consumer_count == 0
        assert queue2.declaration_result.message_count == 1

    async def test_queue_empty_exception(
        self, channel: aio_pika.Channel, add_cleanup: Callable, declare_queue,
    ):
        queue_name = get_random_name("test_get_on_empty_queue")

        queue = await declare_queue(queue_name, auto_delete=True)
        add_cleanup(queue.delete)

        with pytest.raises(aio_pika.exceptions.QueueEmpty):
            await queue.get(timeout=5)

        await channel.default_exchange.publish(
            Message(b"test"), queue_name,
        )

        message = await queue.get(timeout=5)
        assert message.body == b"test"

        # test again for #110
        with pytest.raises(aio_pika.exceptions.QueueEmpty):
            await queue.get(timeout=5)

    async def test_queue_empty_fail_false(
        self, channel: aio_pika.Channel, declare_queue,
    ):
        queue_name = get_random_name("test_get_on_empty_queue")
        queue = await declare_queue(queue_name, auto_delete=True)

        result = await queue.get(fail=False)
        assert result is None

    async def test_message_nack(
        self, channel: aio_pika.Channel, declare_queue,
    ):
        queue_name = get_random_name("test_nack_queue")
        body = uuid.uuid4().bytes
        queue = await declare_queue(queue_name, auto_delete=True)

        await channel.default_exchange.publish(
            Message(body=body), routing_key=queue_name,
        )

        message = await queue.get()  # type: aio_pika.IncomingMessage

        assert message.body == body
        await message.nack(requeue=True)

        message = await queue.get()

        assert message.redelivered
        assert message.body == body
        await message.ack()

    async def test_on_return_raises(self, connection: aio_pika.Connection):
        queue_name = get_random_name("test_on_return_raises")
        body = uuid.uuid4().bytes

        with pytest.raises(RuntimeError):
            await connection.channel(
                publisher_confirms=False, on_return_raises=True,
            )

        channel = await connection.channel(
            publisher_confirms=True, on_return_raises=True,
        )

        for _ in range(100):
            with pytest.raises(aio_pika.exceptions.DeliveryError):
                await channel.default_exchange.publish(
                    Message(body=body), routing_key=queue_name,
                )

    async def test_transaction_when_publisher_confirms_error(
        self, connection: aio_pika.Connection,
    ):
        async with connection.channel(publisher_confirms=True) as channel:
            with pytest.raises(RuntimeError):
                channel.transaction()

    async def test_transaction_simple_commit(
        self, connection: aio_pika.Connection,
    ):
        async with connection.channel(publisher_confirms=False) as channel:
            tx = channel.transaction()
            await tx.select()
            await tx.commit()

    async def test_transaction_simple_rollback(
        self, connection: aio_pika.Connection,
    ):
        async with connection.channel(publisher_confirms=False) as channel:
            tx = channel.transaction()
            await tx.select()
            await tx.rollback()

    async def test_transaction_simple_async_commit(
        self, connection: aio_pika.Connection,
    ):
        async with connection.channel(publisher_confirms=False) as channel:
            async with channel.transaction():
                pass

    async def test_transaction_simple_async_rollback(
        self, connection: aio_pika.Connection,
    ):
        async with connection.channel(publisher_confirms=False) as channel:
            with pytest.raises(ValueError):
                async with channel.transaction():
                    raise ValueError

    async def test_async_for_queue(
        self, event_loop, connection, declare_queue,
    ):
        channel2 = await self.create_channel(connection)

        queue = await declare_queue(
            get_random_name("queue", "is_async", "for"),
            auto_delete=True,
            channel=channel2,
        )

        messages = 100

        async def publisher():
            channel1 = await self.create_channel(connection)

            for i in range(messages):
                await channel1.default_exchange.publish(
                    Message(body=str(i).encode()), routing_key=queue.name,
                )

        event_loop.create_task(publisher())

        count = 0
        data = list()

        async for message in queue:
            async with message.process():
                count += 1
                data.append(message.body)

            if count >= messages:
                break

        assert data == list(map(lambda x: str(x).encode(), range(messages)))

    async def test_async_for_queue_context(
        self, event_loop, connection, declare_queue,
    ) -> None:
        channel2 = await self.create_channel(connection)

        queue = await declare_queue(
            get_random_name("queue", "is_async", "for"),
            auto_delete=True,
            channel=channel2,
        )

        messages: asyncio.Queue[bytes] = asyncio.Queue(100)
        condition = asyncio.Condition()

        async def publisher() -> None:
            channel1 = await self.create_channel(connection)

            for i in range(messages.maxsize):
                body = str(i).encode()
                await messages.put(body)
                await channel1.default_exchange.publish(
                    Message(body=body), routing_key=queue.name,
                )

        async def consumer() -> None:
            async with queue.iterator() as queue_iterator:
                async for message in queue_iterator:
                    async with message.process():
                        async with condition:
                            data.append(message.body)
                            messages.task_done()
                            condition.notify()

        async def application_stop_request() -> None:
            async with condition:
                await condition.wait_for(messages.full)
            await messages.join()
            await asyncio.sleep(1)
            await connection.close()

        p = event_loop.create_task(publisher())
        c = event_loop.create_task(consumer())
        asr = event_loop.create_task(application_stop_request())

        data: List[bytes] = list()

        await asyncio.gather(p, c, asr)

        assert data == list(
            map(lambda x: str(x).encode(), range(messages.maxsize))
        )

    async def test_async_with_connection(
        self, create_connection: Callable,
        connection, event_loop, declare_queue,
    ):
        async with await create_connection() as connection:

            channel = await self.create_channel(connection)

            queue = await declare_queue(
                get_random_name("queue", "is_async", "for"),
                auto_delete=True,
                channel=channel,
            )

            condition = asyncio.Condition()
            messages: asyncio.Queue[bytes] = asyncio.Queue(
                100
            )

            async def publisher():
                channel1 = await self.create_channel(connection)

                for i in range(messages.maxsize):
                    body = str(i).encode()
                    await messages.put(body)
                    await channel1.default_exchange.publish(
                        Message(body=body), routing_key=queue.name,
                    )

            data = list()

            async def consume_loop():
                async with queue.iterator() as queue_iterator:
                    async for message in queue_iterator:
                        async with message.process():
                            async with condition:
                                data.append(message.body)
                                condition.notify()
                                messages.task_done()

            async def application_close_request():
                async with condition:
                    await condition.wait_for(messages.full)
                await messages.join()
                await asyncio.sleep(1)
                await connection.close()

            p = event_loop.create_task(publisher())
            cl = event_loop.create_task(consume_loop())
            acr = event_loop.create_task(application_close_request())

            await asyncio.gather(p, cl, acr)

            assert data == list(
                map(lambda x: str(x).encode(), range(messages.maxsize)),
            )

        assert channel.is_closed

    async def test_async_with_channel(self, connection: aio_pika.Connection):
        async with self.create_channel(connection) as channel:
            assert isinstance(channel, Channel)

        assert channel.is_closed

    async def test_delivery_fail(
        self, channel: aio_pika.Channel, declare_queue,
    ):
        queue = await declare_queue(
            exclusive=True,
            arguments={"x-max-length": 1, "x-overflow": "reject-publish"},
            auto_delete=True,
        )

        await channel.default_exchange.publish(
            aio_pika.Message(body=b"queue me"), routing_key=queue.name,
        )

        with pytest.raises(DeliveryError):
            for _ in range(10):
                await channel.default_exchange.publish(
                    aio_pika.Message(body=b"reject me"),
                    routing_key=queue.name,
                )

    async def test_channel_locked_resource(
        self, connection, declare_queue, add_cleanup: Callable,
    ):
        ch1 = await self.create_channel(connection)
        ch2 = await self.create_channel(connection)

        qname = get_random_name("channel", "locked", "resource")

        q1 = await declare_queue(
            qname, exclusive=True, channel=ch1, cleanup=False,
        )
        add_cleanup(q1.delete)

        tag = await q1.consume(print, exclusive=True)
        add_cleanup(q1.cancel, tag)

        with pytest.raises(aiormq.exceptions.ChannelAccessRefused):
            q2 = await declare_queue(
                qname, exclusive=True, channel=ch2, cleanup=False,
            )
            await q2.consume(print, exclusive=True)

    async def test_queue_iterator_close_was_called_twice(
        self, create_connection: Callable, event_loop, declare_queue,
    ):
        event = asyncio.Event()

        queue_name = get_random_name()
        iterator: QueueIterator = None  # type: ignore[assignment]

        async def task_inner():
            nonlocal iterator
            connection = await create_connection()

            async with connection:
                channel = await self.create_channel(connection)

                queue = await declare_queue(
                    queue_name, channel=channel, cleanup=False,
                )

                async with queue.iterator() as iterator:
                    event.set()

                    async for message in iterator:
                        async with message.process():
                            pytest.fail("who sent this message?")

        task = event_loop.create_task(task_inner())

        await event.wait()
        await iterator.close()
        await task

    async def test_queue_iterator_close_with_noack(
        self,
        create_connection: Callable,
        event_loop,
        add_cleanup: Callable,
        declare_queue,
    ):
        messages: asyncio.Queue = asyncio.Queue()
        queue_name = get_random_name("test_queue")
        body = get_random_name("test_body").encode()

        async def task_inner():
            connection = await create_connection()
            add_cleanup(connection.close)

            async with connection:
                channel = await self.create_channel(connection)

                queue = await declare_queue(
                    queue_name, channel=channel, cleanup=False, passive=True,
                )

                async with queue.iterator(no_ack=True) as q:
                    async for message in q:
                        await messages.put(message)
                        return

        async with await create_connection() as connection:
            channel = await self.create_channel(connection)

            queue = await declare_queue(
                queue_name, channel=channel, cleanup=False,
            )

            try:
                await channel.default_exchange.publish(
                    Message(body), routing_key=queue_name,
                )

                task = event_loop.create_task(task_inner())

                message = await messages.get()

                assert message
                assert message.body == body

            finally:
                await task
                await queue.delete()

    async def test_queue_iterator_throws_cancelled_error(
        self,
        create_connection: Callable,
        event_loop,
        add_cleanup: Callable,
        declare_queue,
    ):
        event_loop.set_debug(True)
        queue_name = get_random_name("test_queue")

        connection = await create_connection()

        async with connection:
            channel = await self.create_channel(connection)

            queue = await channel.declare_queue(
                queue_name,
            )

            iterator = queue.iterator()
            task = event_loop.create_task(iterator.__anext__())
            done, pending = await asyncio.wait({task}, timeout=1)
            assert not done
            task.cancel()

            with pytest.raises(asyncio.CancelledError):
                await task

    async def test_queue_iterator_throws_timeout_error(
        self,
        create_connection: Callable,
        event_loop,
        add_cleanup: Callable,
        declare_queue,
    ):
        event_loop.set_debug(True)
        queue_name = get_random_name("test_queue")

        connection = await create_connection()

        async with connection:
            channel = await self.create_channel(connection)

            queue = await channel.declare_queue(
                queue_name,
            )

            iterator = queue.iterator(timeout=1)
            task = event_loop.create_task(iterator.__anext__())
            done, pending = await asyncio.wait({task}, timeout=5)
            assert done

            with pytest.raises(asyncio.TimeoutError):
                await task

    async def test_passive_for_exchange(
        self, declare_exchange: Callable, connection, add_cleanup: Callable,
    ):
        name = get_random_name("passive", "exchange")

        ch1 = await self.create_channel(connection)
        ch2 = await self.create_channel(connection)
        ch3 = await self.create_channel(connection)

        with pytest.raises(aio_pika.exceptions.ChannelNotFoundEntity):
            await declare_exchange(name, passive=True, channel=ch1)

        exchange = await declare_exchange(name, auto_delete=True, channel=ch2)
        exchange_passive = await declare_exchange(
            name, passive=True, channel=ch3,
        )

        assert exchange.name == exchange_passive.name

    async def test_passive_queue(
        self, declare_queue: Callable, connection: aio_pika.Connection,
    ):
        name = get_random_name("passive", "queue")

        ch1 = await self.create_channel(connection)
        ch2 = await self.create_channel(connection)
        ch3 = await self.create_channel(connection)

        with pytest.raises(aio_pika.exceptions.ChannelNotFoundEntity):
            await declare_queue(name, passive=True, channel=ch1)

        queue = await declare_queue(name, auto_delete=True, channel=ch2)
        queue_passive = await declare_queue(name, passive=True, channel=ch3)

        assert queue.name == queue_passive.name

    async def test_get_exchange(self, connection, declare_exchange):
        channel = await self.create_channel(connection)
        name = get_random_name("passive", "exchange")

        with pytest.raises(aio_pika.exceptions.ChannelNotFoundEntity):
            await channel.get_exchange(name)

        channel = await self.create_channel(connection)
        exchange = await declare_exchange(
            name, auto_delete=True, channel=channel,
        )
        exchange_passive = await channel.get_exchange(name)

        assert exchange.name == exchange_passive.name

    async def test_get_queue(self, connection, declare_queue):
        channel = await self.create_channel(connection)
        name = get_random_name("passive", "queue")

        with pytest.raises(aio_pika.exceptions.ChannelNotFoundEntity):
            await channel.get_queue(name)

        channel = await self.create_channel(connection)
        queue = await declare_queue(name, auto_delete=True, channel=channel)
        queue_passive = await channel.get_queue(name)

        assert queue.name, queue_passive.name

    @pytest.mark.skip(reason="temporary skip")
    async def test_channel_blocking_timeout(self, connection):
        channel = await connection.channel()
        close_reasons = []
        close_event = asyncio.Event()

        def on_done(*args):
            close_reasons.append(args)
            close_event.set()
            return

        channel.close_callbacks.add(on_done)

        async def run():
            await channel.set_qos(1)
            time.sleep(1)
            await channel.set_qos(0)

        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(run(), timeout=0.2)

        await close_event.wait()

        assert channel.is_closed

        # Ensure close callback has been called
        assert close_reasons

        with pytest.raises(RuntimeError):
            await channel.set_qos(10)

    async def test_heartbeat_disabling(
        self, event_loop, amqp_url: URL, connection_fabric,
    ):
        url = amqp_url.update_query(heartbeat=0)
        connection: AbstractConnection = await connection_fabric(url)

        transport = connection.transport
        assert transport
        heartbeat = transport.connection.connection_tune.heartbeat

        async with connection:
            assert heartbeat == 0

    async def test_non_acked_messages_are_redelivered_to_queue(
        self,
        channel: aio_pika.Channel,
        declare_queue: Callable,
        declare_exchange: Callable,
    ):
        queue_name = get_random_name("test_connection")
        routing_key = get_random_name()

        exchange = await declare_exchange(
            "direct", auto_delete=True, channel=channel,
        )

        queue: AbstractQueue = await declare_queue(
            queue_name, auto_delete=False, channel=channel,
        )

        await queue.bind(exchange, routing_key)

        # Publish 5 messages to queue
        all_bodies = []
        for _ in range(0, 5):
            body = bytes(shortuuid.uuid(), "utf-8")
            all_bodies.append(body)

            assert await exchange.publish(Message(body), routing_key)

        # Create a subscription but only process first message
        async with queue.iterator() as queue_iterator:
            first_message = await queue_iterator.__anext__()
            async with first_message.process():
                assert first_message.body == all_bodies[0]

        # Confirm other messages are still in queue
        for i in range(1, 5):
            incoming_message = await queue.get(timeout=5)
            await incoming_message.ack()

            assert incoming_message.body == all_bodies[i]

        # Check if the queue is now empty
        assert await queue.get(fail=False, timeout=.5) is None

        # Cleanup, delete the queue
        await queue.delete()

    async def test_regression_only_messages_cancelled_subscription_are_nacked(
        self,
        channel: aio_pika.Channel,
        declare_queue: Callable,
        declare_exchange: Callable,
    ):
        queue_name1 = get_random_name("test_queue")
        queue_name2 = get_random_name("test_queue")
        routing_key1 = get_random_name()
        routing_key2 = get_random_name()

        exchange = await declare_exchange(
            "direct", auto_delete=True, channel=channel,
        )

        queue1: AbstractQueue = await declare_queue(
            queue_name1, auto_delete=False, channel=channel,
        )
        queue2: AbstractQueue = await declare_queue(
            queue_name2, auto_delete=False, channel=channel,
        )

        await queue1.bind(exchange, routing_key1)
        await queue2.bind(exchange, routing_key2)

        # Publish 5 messages to queue 1
        all_bodies1 = []
        for _ in range(0, 5):
            body = bytes(shortuuid.uuid(), "utf-8")
            all_bodies1.append(body)

            assert await exchange.publish(Message(body), routing_key1)

        # Publish 5 messages to queue 2
        all_bodies2 = []
        for _ in range(0, 5):
            body = bytes(shortuuid.uuid(), "utf-8")
            all_bodies2.append(body)

            assert await exchange.publish(Message(body), routing_key2)

        # Create a subscription to both queues but only process first message
        queue_iterator1 = await queue1.iterator().__aenter__()
        queue_iterator2 = await queue2.iterator().__aenter__()

        first_message1 = await queue_iterator1.__anext__()
        async with first_message1.process():
            assert first_message1.body == all_bodies1[0]

        first_message2 = await queue_iterator2.__anext__()
        async with first_message2.process():
            assert first_message2.body == all_bodies2[0]
        #  The order of exit here is important.
        #    Subscription to queue 1 is received first then to 2.
        #    Therefore, the delivery tags of subscription to queue 2 will be
        #    higher.
        #    So first we cancel the subscription to 2, to test if we
        #    accidentally also nacked the messages of queue 1. Then we cancel
        #    subscription to queue 1 to test.

        await queue_iterator2.__aexit__(None, None, None)
        # To test if the wrong messages are nacked by stopping subscription to
        # queue 2, we ack a message received from queue 1. If it was nacked,
        # RabbitMQ will throw an exception.
        second_message1 = await queue_iterator1.__anext__()
        async with second_message1.process():
            assert second_message1.body == all_bodies1[1]

        await queue_iterator1.__aexit__(None, None, None)

        # Confirm other messages are still in queue
        for i in range(2, 5):
            incoming_message = await queue1.get(timeout=5)
            await incoming_message.ack()
            assert incoming_message.body == all_bodies1[i]

        for i in range(1, 5):
            incoming_message = await queue2.get(timeout=5)
            await incoming_message.ack()
            assert incoming_message.body == all_bodies2[i]

        # Check if the queue is now empty
        assert await queue1.get(fail=False, timeout=.5) is None
        assert await queue2.get(fail=False, timeout=.5) is None

        # Cleanup, delete the queue
        await queue1.delete()
        await queue2.delete()


class TestCaseAmqpNoConfirms(TestCaseAmqp):
    @staticmethod
    def create_channel(connection: aio_pika.Connection):
        return connection.channel(publisher_confirms=False)


class TestCaseAmqpWithConfirms(TestCaseAmqpBase):
    @staticmethod
    def create_channel(connection: aio_pika.Connection):
        return connection.channel(publisher_confirms=True)

    @pytest.mark.skip(reason="Have to find another way to close connection")
    async def test_connection_close(
        self, connection: aio_pika.Connection, declare_exchange: Callable,
    ):
        routing_key = get_random_name()

        channel = await self.create_channel(connection)
        exchange = await declare_exchange(
            "direct", auto_delete=True, channel=channel,
        )

        try:
            with pytest.raises(aio_pika.exceptions.ChannelPreconditionFailed):
                msg = Message(bytes(shortuuid.uuid(), "utf-8"))
                msg.delivery_mode = 8       # type: ignore

                await exchange.publish(msg, routing_key)

            channel = await self.create_channel(connection)
            exchange = await declare_exchange(
                "direct", auto_delete=True, channel=channel,
            )
        finally:
            await exchange.delete()

    async def test_basic_return(
        self, connection: aio_pika.Connection, event_loop,
    ):
        channel = await self.create_channel(connection)

        f = event_loop.create_future()

        def handler(channel, message: ReturnedMessage):
            f.set_result(message)

        channel.return_callbacks.add(handler)

        body = bytes(shortuuid.uuid(), "utf-8")

        await channel.default_exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            get_random_name("test_basic_return"),
        )

        returned = await f

        assert returned.body == body

        # handler with exception
        f = event_loop.create_future()

        await channel.close()

        channel = await self.create_channel(connection)

        def bad_handler(
            channel: aio_pika.abc.AbstractChannel,
            message: aio_pika.message.IncomingMessage,
        ):
            try:
                raise ValueError
            finally:
                f.set_result(message)

        channel.return_callbacks.add(bad_handler)

        body = bytes(shortuuid.uuid(), "utf-8")

        await channel.default_exchange.publish(
            Message(body, content_type="text/plain", headers={"foo": "bar"}),
            get_random_name("test_basic_return"),
        )

        returned = await f

        assert returned.body == body



import time
from copy import copy
from datetime import datetime, timezone
from typing import List, Tuple

import shortuuid

from aio_pika import DeliveryMode, Message
from aio_pika.abc import FieldValue, HeadersType, MessageInfo


def test_message_copy():
    msg1 = Message(
        bytes(shortuuid.uuid(), "utf-8"),
        content_type="application/json",
        content_encoding="text",
        timestamp=datetime(2000, 1, 1),
        headers={"h1": "v1", "h2": "v2"},
    )
    msg2 = copy(msg1)

    msg1.lock()

    assert not msg2.locked


def test_message_info():
    body = bytes(shortuuid.uuid(), "utf-8")

    info = MessageInfo(
        app_id="test",
        body_size=len(body),
        cluster_id=None,
        consumer_tag=None,
        content_encoding="text",
        content_type="application/json",
        correlation_id="1",
        delivery_mode=DeliveryMode.PERSISTENT,
        delivery_tag=None,
        exchange=None,
        expiration=1.5,
        headers={"foo": "bar"},
        message_id=shortuuid.uuid(),
        priority=0,
        redelivered=None,
        reply_to="test",
        routing_key=None,
        timestamp=datetime.fromtimestamp(int(time.time()), tz=timezone.utc),
        type="0",
        user_id="guest",
    )

    msg = Message(
        body=body,
        headers={"foo": "bar"},
        content_type="application/json",
        content_encoding="text",
        delivery_mode=DeliveryMode.PERSISTENT,
        priority=0,
        correlation_id="1",
        reply_to="test",
        expiration=1.5,
        message_id=info["message_id"],
        timestamp=info["timestamp"],
        type="0",
        user_id="guest",
        app_id="test",
    )

    assert info == msg.info()


def test_headers_setter():
    data: HeadersType = {"foo": "bar"}
    data_expected = {"foo": "bar"}

    msg = Message(b"", headers={"bar": "baz"})
    msg.headers = data

    assert msg.headers == data_expected


def test_headers_content():
    data: Tuple[List[FieldValue], ...] = (
        [42, 42],
        [b"foo", b"foo"],
        [b"\00", b"\00"],
    )

    for src, value in data:
        msg = Message(b"", headers={"value": src})
        assert msg.headers["value"] == value


def test_headers_set():
    msg = Message(b"", headers={"header": "value"})

    data = (
        ["header-1", 42,  42],
        ["header-2", b"foo", b"foo"],
        ["header-3", b"\00", b"\00"],
        ["header-4", {"foo": "bar"}, {"foo": "bar"}],
    )

    for name, src, value in data:           # type: ignore
        msg.headers[name] = value           # type: ignore
        assert msg.headers[name] == value   # type: ignore

    assert msg.headers["header"] == "value"



import asyncio
import gzip
import json
import logging
from functools import partial
from types import MappingProxyType
from typing import Any, Awaitable, Mapping, Optional

import aiormq

from aio_pika.abc import (
    AbstractChannel, AbstractExchange, AbstractIncomingMessage, AbstractQueue,
    ConsumerTag, DeliveryMode,
)
from aio_pika.message import Message

from ..tools import create_task, ensure_awaitable
from .base import Base, CallbackType, Proxy, T


log = logging.getLogger(__name__)


class MessageProcessingError(Exception):
    pass


class NackMessage(MessageProcessingError):
    def __init__(self, requeue: bool = False):
        self.requeue = requeue


class RejectMessage(MessageProcessingError):
    def __init__(self, requeue: bool = False):
        self.requeue = requeue


class Worker:
    __slots__ = (
        "queue",
        "consumer_tag",
        "loop",
    )

    def __init__(
        self, queue: AbstractQueue, consumer_tag: ConsumerTag,
        loop: asyncio.AbstractEventLoop,
    ):
        self.queue = queue
        self.consumer_tag = consumer_tag
        self.loop = loop

    def close(self) -> Awaitable[None]:
        """ Cancel subscription to the channel

        :return: :class:`asyncio.Task`
        """

        async def closer() -> None:
            await self.queue.cancel(self.consumer_tag)

        return create_task(closer)


class Master(Base):
    __slots__ = (
        "channel",
        "loop",
        "proxy",
        "_requeue",
        "_reject_on_redelivered",
    )

    DELIVERY_MODE = DeliveryMode.PERSISTENT

    __doc__ = """
    Implements Master/Worker pattern.
    Usage example:

    `worker.py` ::

        master = Master(channel)
        worker = await master.create_worker('test_worker', lambda x: print(x))

    `master.py` ::

        master = Master(channel)
        await master.proxy.test_worker('foo')
    """

    def __init__(
        self,
        channel: AbstractChannel,
        requeue: bool = True,
        reject_on_redelivered: bool = False,
    ):
        """ Creates a new :class:`Master` instance.

        :param channel: Initialized instance of :class:`aio_pika.Channel`
        """
        self.channel: AbstractChannel = channel
        self.loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        self.proxy = Proxy(self.create_task)

        self.channel.return_callbacks.add(self.on_message_returned)

        self._requeue = requeue
        self._reject_on_redelivered = reject_on_redelivered

    @property
    def exchange(self) -> AbstractExchange:
        return self.channel.default_exchange

    @staticmethod
    def on_message_returned(
        channel: Optional[AbstractChannel],
        message: AbstractIncomingMessage,
    ) -> None:
        log.warning(
            "Message returned. Probably destination queue does not exists: %r",
            message,
        )

    def serialize(self, data: Any) -> bytes:
        """ Serialize data to the bytes.
        Uses `pickle` by default.
        You should overlap this method when you want to change serializer

        :param data: Data which will be serialized
        :returns: bytes
        """
        return super().serialize(data)

    def deserialize(self, data: bytes) -> Any:
        """ Deserialize data from bytes.
        Uses `pickle` by default.
        You should overlap this method when you want to change serializer

        :param data: Data which will be deserialized
        :returns: :class:`Any`
        """
        return super().deserialize(data)

    @classmethod
    async def execute(
        cls, func: CallbackType, kwargs: Any,
    ) -> T:
        kwargs = kwargs or {}

        if not isinstance(kwargs, dict):
            logging.error("Bad kwargs %r received for the %r", kwargs, func)
            raise RejectMessage(requeue=False)

        return await func(**kwargs)

    async def on_message(
        self, func: CallbackType,
        message: AbstractIncomingMessage,
    ) -> None:
        async with message.process(
            requeue=self._requeue,
            reject_on_redelivered=self._reject_on_redelivered,
            ignore_processed=True,
        ):
            try:
                await self.execute(func, self.deserialize(message.body))
            except RejectMessage as e:
                await message.reject(requeue=e.requeue)
            except NackMessage as e:
                await message.nack(requeue=e.requeue)

    async def create_queue(
        self, queue_name: str, **kwargs: Any,
    ) -> AbstractQueue:
        return await self.channel.declare_queue(queue_name, **kwargs)

    async def create_worker(
        self, queue_name: str,
        func: CallbackType,
        **kwargs: Any,
    ) -> Worker:
        """ Creates a new :class:`Worker` instance. """
        queue = await self.create_queue(queue_name, **kwargs)
        consumer_tag = await queue.consume(
            partial(self.on_message, ensure_awaitable(func)),
        )

        return Worker(queue, consumer_tag, self.loop)

    async def create_task(
        self, channel_name: str,
        kwargs: Mapping[str, Any] = MappingProxyType({}),
        **message_kwargs: Any,
    ) -> Optional[aiormq.abc.ConfirmationFrameType]:

        """ Creates a new task for the worker """
        message = Message(
            body=self.serialize(kwargs),
            content_type=self.CONTENT_TYPE,
            delivery_mode=self.DELIVERY_MODE,
            **message_kwargs,
        )

        return await self.exchange.publish(
            message, channel_name, mandatory=True,
        )


class JsonMaster(Master):
    SERIALIZER = json
    CONTENT_TYPE = "application/json"

    def serialize(self, data: Any) -> bytes:
        return self.SERIALIZER.dumps(data, ensure_ascii=False).encode()


class CompressedJsonMaster(Master):
    SERIALIZER = json
    CONTENT_TYPE = "application/json;compression=gzip"
    COMPRESS_LEVEL = 6

    def serialize(self, data: Any) -> bytes:
        return gzip.compress(
            self.SERIALIZER.dumps(data, ensure_ascii=False).encode(),
            compresslevel=self.COMPRESS_LEVEL,
        )

    def deserialize(self, data: bytes) -> Any:
        return self.SERIALIZER.loads(gzip.decompress(data))



import asyncio
import uuid
import warnings
from typing import Any, Awaitable, Callable, Dict, Optional, Tuple, Union

import aiormq
from aiormq import ChannelInvalidStateError
from pamqp.common import Arguments

from .abc import (
    AbstractChannel, AbstractExchange, AbstractIncomingMessage,
    AbstractQueueIterator, AbstractRobustQueue, ConsumerTag, TimeoutType,
)
from .exchange import ExchangeParamType
from .log import get_logger
from .queue import Queue, QueueIterator


log = get_logger(__name__)


class RobustQueue(Queue, AbstractRobustQueue):
    __slots__ = ("_consumers", "_bindings")

    _consumers: Dict[ConsumerTag, Dict[str, Any]]
    _bindings: Dict[Tuple[Union[AbstractExchange, str], str], Dict[str, Any]]

    def __init__(
        self,
        channel: AbstractChannel,
        name: Optional[str],
        durable: bool = False,
        exclusive: bool = False,
        auto_delete: bool = False,
        arguments: Arguments = None,
        passive: bool = False,
    ):

        super().__init__(
            channel=channel,
            name=name or f"amq_{uuid.uuid4().hex}",
            durable=durable,
            exclusive=exclusive,
            auto_delete=auto_delete,
            arguments=arguments,
            passive=passive,
        )

        self._consumers = {}
        self._bindings = {}

    async def restore(self, channel: Any = None) -> None:
        if channel is not None:
            warnings.warn(
                "Channel argument will be ignored because you "
                "don't need to pass this anymore.",
                DeprecationWarning,
            )

        await self.declare()
        bindings = tuple(self._bindings.items())
        consumers = tuple(self._consumers.items())

        for (exchange, routing_key), kwargs in bindings:
            await self.bind(exchange, routing_key, **kwargs)

        for consumer_tag, kwargs in consumers:
            await self.consume(consumer_tag=consumer_tag, **kwargs)

    async def bind(
        self,
        exchange: ExchangeParamType,
        routing_key: Optional[str] = None,
        *,
        arguments: Arguments = None,
        timeout: TimeoutType = None,
        robust: bool = True,
    ) -> aiormq.spec.Queue.BindOk:
        if routing_key is None:
            routing_key = self.name

        result = await super().bind(
            exchange=exchange, routing_key=routing_key,
            arguments=arguments, timeout=timeout,
        )

        if robust:
            self._bindings[(exchange, routing_key)] = dict(
                arguments=arguments,
            )

        return result

    async def unbind(
        self,
        exchange: ExchangeParamType,
        routing_key: Optional[str] = None,
        arguments: Arguments = None,
        timeout: TimeoutType = None,
    ) -> aiormq.spec.Queue.UnbindOk:
        if routing_key is None:
            routing_key = self.name

        result = await super().unbind(
            exchange, routing_key, arguments, timeout,
        )
        self._bindings.pop((exchange, routing_key), None)

        return result

    async def consume(
        self,
        callback: Callable[[AbstractIncomingMessage], Awaitable[Any]],
        no_ack: bool = False,
        exclusive: bool = False,
        arguments: Arguments = None,
        consumer_tag: Optional[ConsumerTag] = None,
        timeout: TimeoutType = None,
        robust: bool = True,
    ) -> ConsumerTag:
        consumer_tag = await super().consume(
            consumer_tag=consumer_tag,
            timeout=timeout,
            callback=callback,
            no_ack=no_ack,
            exclusive=exclusive,
            arguments=arguments,
        )

        if robust:
            self._consumers[consumer_tag] = dict(
                callback=callback,
                no_ack=no_ack,
                exclusive=exclusive,
                arguments=arguments,
            )

        return consumer_tag

    async def cancel(
        self,
        consumer_tag: ConsumerTag,
        timeout: TimeoutType = None,
        nowait: bool = False,
    ) -> aiormq.spec.Basic.CancelOk:
        result = await super().cancel(consumer_tag, timeout, nowait)
        self._consumers.pop(consumer_tag, None)
        return result

    def iterator(self, **kwargs: Any) -> AbstractQueueIterator:
        return RobustQueueIterator(self, **kwargs)


class RobustQueueIterator(QueueIterator):
    """Queue iterator that survives channel reconnection.

    This iterator handles channel disconnection/reconnection gracefully
    by waiting for channel restoration instead of raising StopAsyncIteration.
    """

    RETRY_DELAY: float = 0.5

    def __init__(self, queue: Queue, **kwargs: Any):
        super().__init__(queue, **kwargs)

        # Remove close callback to survive reconnection
        self._amqp_queue.close_callbacks.discard(self._set_closed)

        # But listen to connection close to stop iteration when
        # connection is intentionally closed
        channel = self._amqp_queue.channel
        if hasattr(channel, "_connection"):
            connection = channel._connection
            connection.closed().add_done_callback(self._on_connection_closed)

    def _on_connection_closed(self, _: asyncio.Future) -> None:
        """Handle connection closed - set _closed to stop iteration."""
        if not self._closed.done():
            self._closed.set_result(True)
        self._message_or_closed.set()

    async def consume(self) -> None:
        """Consume with retry on channel errors.

        Waits for channel to be ready before consuming, with backoff delay
        between retries to prevent CPU spinning during reconnection.
        """
        while True:
            try:
                # Wait for channel to be fully ready before consuming
                channel = self._amqp_queue.channel
                if hasattr(channel, "ready"):
                    await channel.ready()

                return await super().consume()
            except ChannelInvalidStateError:
                log.debug(
                    "Channel invalid state in %r, waiting for restoration",
                    self,
                )
                # Backoff to prevent CPU spinning during reconnection
                await asyncio.sleep(self.RETRY_DELAY)

    async def __anext__(self) -> AbstractIncomingMessage:
        """Get next message, handling reconnection gracefully.

        During reconnection, the queue may be empty and _message_or_closed
        may be set due to channel closure. In robust mode, we wait for
        reconnection instead of raising StopAsyncIteration.
        """
        while True:
            # Check if explicitly closed
            if self._closed.done():
                raise StopAsyncIteration

            if not hasattr(self, "_consumer_tag"):
                await self.consume()

            timeout: Optional[float] = self._consume_kwargs.get("timeout")

            if not self._message_or_closed.is_set():
                coroutine: Awaitable[Any] = self._message_or_closed.wait()
                if timeout is not None and timeout > 0:
                    coroutine = asyncio.wait_for(coroutine, timeout=timeout)

                try:
                    await coroutine
                except (asyncio.TimeoutError, asyncio.CancelledError):
                    # Handle timeout same as parent class
                    if timeout is not None:
                        timeout_val = (
                            timeout
                            if timeout > 0
                            else self.DEFAULT_CLOSE_TIMEOUT
                        )
                        log.info(
                            "%r closing with timeout %d seconds",
                            self, timeout_val,
                        )

                    task = asyncio.create_task(self.close())
                    close_coro: Awaitable[Any] = task
                    if timeout is not None:
                        close_coro = asyncio.wait_for(
                            asyncio.shield(task),
                            timeout=timeout_val,
                        )

                    try:
                        await close_coro
                    except asyncio.TimeoutError:
                        self._QueueIterator__closing = task

                    raise

            # Check queue for messages
            if not self._queue.empty():
                msg = self._queue.get_nowait()

                if (
                    self._queue.empty()
                    and not self._amqp_queue.channel.is_closed
                    and not self._closed.done()
                ):
                    self._message_or_closed.clear()

                return msg

            # Queue is empty - check if this is a reconnection scenario
            channel = self._amqp_queue.channel
            if hasattr(channel, "ready"):
                # This is a RobustChannel - check if connection is still alive
                if hasattr(channel, "_connection"):
                    connection = channel._connection
                    # Only wait for reconnection if connection wasn't
                    # intentionally closed
                    if not connection.is_closed and not connection.close_called:
                        # Connection is alive, channel is being restored
                        log.debug(
                            "%r queue empty during channel restoration, "
                            "waiting for reconnection",
                            self,
                        )

                        # Clear the event and wait for channel restoration
                        self._message_or_closed.clear()

                        try:
                            # Wait for channel to become ready
                            await asyncio.wait_for(
                                channel.ready(), timeout=60.0,
                            )

                            # Re-establish consumer if needed
                            if not hasattr(self, "_consumer_tag"):
                                await self.consume()

                            # Continue loop to wait for new messages
                            continue
                        except asyncio.TimeoutError:
                            log.error(
                                "%r timeout waiting for channel reconnection",
                                self,
                            )
                            raise StopAsyncIteration

            # Truly empty and not reconnecting - stop iteration
            raise StopAsyncIteration


__all__ = ("RobustQueue",)



import asyncio
import logging
import warnings
from functools import partial

import pytest

import aio_pika
from aio_pika import Message
from aio_pika.exceptions import MessageProcessError
from aio_pika.message import IncomingMessage
from aio_pika.patterns.rpc import RPC
from aio_pika.patterns.rpc import log as rpc_logger
from tests import get_random_name


async def rpc_func(*, foo, bar):
    assert not foo
    assert not bar

    return {"foo": "bar"}


async def rpc_func2(*, foo, bar):
    assert not foo
    assert not bar

    return {"foo": "bar2"}


class TestCase:
    async def test_simple(self, channel: aio_pika.Channel):
        rpc = await RPC.create(channel, auto_delete=True)

        await rpc.register("test.rpc", rpc_func, auto_delete=True)

        result = await rpc.proxy.test.rpc(foo=None, bar=None)
        assert result == {"foo": "bar"}

        await rpc.unregister(rpc_func)
        await rpc.close()

        # Close already closed
        await rpc.close()

    async def test_error(self, channel: aio_pika.Channel):
        rpc = await RPC.create(channel, auto_delete=True)

        await rpc.register("test.rpc", rpc_func, auto_delete=True)

        with pytest.raises(AssertionError):
            await rpc.proxy.test.rpc(foo=True, bar=None)

        await rpc.unregister(rpc_func)
        await rpc.close()

    async def test_unroutable(self, channel: aio_pika.Channel):
        rpc = await RPC.create(channel, auto_delete=True)

        await rpc.register("test.rpc", rpc_func, auto_delete=True)

        with pytest.raises(MessageProcessError):
            await rpc.proxy.unroutable()

        await rpc.unregister(rpc_func)
        await rpc.close()

    async def test_timed_out(self, channel: aio_pika.Channel):
        rpc = await RPC.create(channel, auto_delete=True)

        await rpc.register("test.rpc", rpc_func, auto_delete=True)

        await channel.declare_queue(
            "test.timed_out",
            auto_delete=True,
            arguments={"x-dead-letter-exchange": RPC.DLX_NAME},
        )

        with pytest.raises(asyncio.TimeoutError):
            await rpc.call("test.timed_out", expiration=1)

        await rpc.unregister(rpc_func)
        await rpc.close()

    async def test_close_twice(self, channel: aio_pika.Channel):
        rpc = await RPC.create(channel, auto_delete=True)

        await rpc.close()
        await rpc.close()

    async def test_init_twice(self, channel: aio_pika.Channel):
        rpc = await RPC.create(channel, auto_delete=True)

        await rpc.initialize()

        await rpc.close()

    async def test_send_unknown_message(
        self, channel: aio_pika.Channel, caplog,
    ):
        rpc = await RPC.create(channel, auto_delete=True)

        body = b"test body"
        with caplog.at_level(logging.WARNING, logger=rpc_logger.name):
            await channel.default_exchange.publish(
                Message(body), routing_key=rpc.result_queue.name,
            )

            await asyncio.sleep(0.5)

            for log_record in caplog.records:
                if log_record.levelno == logging.WARNING:
                    break
            else:
                raise pytest.fail("Expected log message")

            incoming = log_record.args[0]

            assert isinstance(incoming, IncomingMessage)
            assert incoming.body == body
            assert (
                "Message without correlation_id was received:"
                in log_record.message
            )

        with caplog.at_level(logging.WARNING, logger=rpc_logger.name):
            await channel.default_exchange.publish(
                Message(body), routing_key="should-returned",
            )

            await asyncio.sleep(0.5)

            for log_record in caplog.records:
                if log_record.levelno == logging.WARNING:
                    break
            else:
                raise pytest.fail("Expected log message")

            incoming = log_record.args[0]
            assert isinstance(incoming, IncomingMessage)
            assert incoming.body == body
            assert (
                "Message without correlation_id was received:"
                in log_record.message
            )

        await rpc.close()

    async def test_close_cancelling(
        self, channel: aio_pika.Channel, event_loop,
    ):
        rpc = await RPC.create(channel, auto_delete=True)

        async def sleeper():
            await asyncio.sleep(60)

        method_name = get_random_name("test", "sleeper")

        await rpc.register(method_name, sleeper, auto_delete=True)

        tasks = set()

        for _ in range(10):
            tasks.add(event_loop.create_task(rpc.call(method_name)))

        await rpc.close()

        logging.info("Waiting for results")
        for task in tasks:
            with pytest.raises(asyncio.CancelledError):
                await task

    async def test_register_twice(self, channel: aio_pika.Channel):
        rpc = await RPC.create(channel, auto_delete=True)

        async def bypass(_: aio_pika.abc.AbstractIncomingMessage):
            return

        await rpc.register("test.sleeper", bypass, auto_delete=True)

        with pytest.raises(RuntimeError):
            await rpc.register(
                "test.sleeper", bypass, auto_delete=True,
            )

        await rpc.register("test.one", rpc_func, auto_delete=True)

        with pytest.raises(RuntimeError):
            await rpc.register("test.two", rpc_func, auto_delete=True)

        await rpc.unregister(rpc_func)
        await rpc.unregister(rpc_func)

        await rpc.close()

    async def test_register_non_coroutine(self, channel: aio_pika.Channel):
        rpc = await RPC.create(channel, auto_delete=True)

        def bypass(_):
            return

        with pytest.deprecated_call():
            await rpc.register(
                "test.non-coroutine",
                bypass,         # type: ignore
                auto_delete=True,
            )

        async def coro(_):
            return

        with pytest.warns(UserWarning) as record:
            warnings.warn("Test", UserWarning)
            await rpc.register(
                "test.coroutine",
                coro,  # type: ignore
                auto_delete=True,
            )

        assert len([w for w in record if w.category is UserWarning]) == 1

        with pytest.warns() as record:
            warnings.warn("Test", UserWarning)
            await rpc.register(
                "test.coroutine_partial",
                partial(partial(coro)),  # type: ignore
                auto_delete=True,
            )

        assert len([w for w in record if w.category is UserWarning]) == 1

    async def test_non_serializable_result(self, channel: aio_pika.Channel):
        rpc = await RPC.create(channel, auto_delete=True)

        async def bad_func():
            async def inner():
                await asyncio.sleep(0)
            return inner()

        await rpc.register(
            "test.not-serializable",
            bad_func,
            auto_delete=True,
        )

        with pytest.raises(TypeError):
            await rpc.call("test.not-serializable")

    async def test_custom_exchange(self, channel: aio_pika.Channel):
        rpc_ex1 = await RPC.create(channel, auto_delete=True, exchange="ex1")
        rpc_ex2 = await RPC.create(channel, auto_delete=True, exchange="ex2")
        rpc_default = await RPC.create(channel, auto_delete=True)

        await rpc_ex1.register("test.rpc", rpc_func, auto_delete=True)
        result = await rpc_ex1.proxy.test.rpc(foo=None, bar=None)
        assert result == {"foo": "bar"}

        with pytest.raises(MessageProcessError):
            await rpc_ex2.proxy.test.rpc(foo=None, bar=None)

        await rpc_ex2.register("test.rpc", rpc_func2, auto_delete=True)
        result = await rpc_ex2.proxy.test.rpc(foo=None, bar=None)
        assert result == {"foo": "bar2"}

        with pytest.raises(MessageProcessError):
            await rpc_default.proxy.test.rpc(foo=None, bar=None)

        await rpc_default.register("test.rpc", rpc_func, auto_delete=True)
        result = await rpc_default.proxy.test.rpc(foo=None, bar=None)
        assert result == {"foo": "bar"}

        await rpc_ex1.unregister(rpc_func)
        await rpc_ex1.close()

        await rpc_ex2.unregister(rpc_func2)
        await rpc_ex2.close()

        await rpc_default.unregister(rpc_func)
        await rpc_default.close()



import asyncio
from collections import Counter

import pytest

from aio_pika.pool import Pool, PoolInstance


@pytest.mark.parametrize("max_size", [50, 10, 5, 1])
async def test_simple(max_size, event_loop):
    counter = 0

    async def create_instance():
        nonlocal counter
        await asyncio.sleep(0)
        counter += 1
        return counter

    pool: Pool = Pool(create_instance, max_size=max_size, loop=event_loop)

    async def getter():
        async with pool.acquire() as instance:
            assert instance > 0
            await asyncio.sleep(1 if counter < max_size else 0)
            return instance, counter

    results = await asyncio.gather(*[getter() for _ in range(200)])

    for instance, total in results:
        assert instance > -1
        assert total > -1

    assert counter == max_size


class TestInstanceBase:
    class Instance(PoolInstance):
        def __init__(self):
            self.closed = False

        async def close(self):
            if self.closed:
                raise RuntimeError

            self.closed = True

    @pytest.fixture
    def instances(self):
        return set()

    @pytest.fixture(params=[50, 40, 30, 20, 10])
    def max_size(self, request):
        return request.param

    @pytest.fixture
    def pool(self, max_size, instances, event_loop):
        async def create_instance():
            obj = TestInstanceBase.Instance()
            instances.add(obj)
            return obj

        return Pool(create_instance, max_size=max_size, loop=event_loop)


class TestInstance(TestInstanceBase):
    async def test_close(self, pool, instances, event_loop, max_size):
        async def getter():
            async with pool.acquire():
                await asyncio.sleep(0.05)

        assert not pool.is_closed
        assert len(instances) == 0

        await asyncio.gather(*[getter() for _ in range(200)])

        assert len(instances) == max_size

        for instance in instances:
            assert not instance.closed

        await pool.close()

        for instance in instances:
            assert instance.closed

        assert pool.is_closed

    async def test_close_context_manager(self, pool, instances):
        async def getter():
            async with pool.acquire():
                await asyncio.sleep(0.05)

        async with pool:
            assert not pool.is_closed

            assert len(instances) == 0

            await asyncio.gather(*[getter() for _ in range(200)])

            assert len(instances) > 1

            for instance in instances:
                assert not instance.closed

            assert not pool.is_closed

        assert pool.is_closed

        for instance in instances:
            assert instance.closed


class TestCaseNoMaxSize(TestInstance):
    async def test_simple(self, pool, event_loop):
        call_count = 200
        counter = 0

        async def getter():
            nonlocal counter
            async with pool.acquire() as instance:
                await asyncio.sleep(1)
                assert isinstance(instance, TestInstanceBase.Instance)
                counter += 1
                return counter

        results = await asyncio.gather(*[getter() for _ in range(call_count)])

        for result in results:
            assert result > -1

        assert counter == call_count


class TestCaseItemReuse(TestInstanceBase):
    @pytest.fixture
    def call_count(self, max_size):
        return max_size * 5

    async def test_simple(self, pool, call_count, instances):
        counter: Counter = Counter()

        async def getter():
            async with pool.acquire() as instance:
                await asyncio.sleep(0.05)
                counter[instance] += 1

        await asyncio.gather(*[getter() for _ in range(call_count)])

        assert sum(counter.values()) == call_count
        assert set(counter) == set(instances)
        assert len(set(counter.values())) == 1
